package com.me.common.ecs;

import com.badlogic.gdx.utils.Array;
import com.me.common.ecs.event.Event;
import com.me.common.ecs.event.Listener;

import java.util.HashMap;
import java.util.Map;

public class Engine {

    private Array<Entity> entities;
    private Array<Entity> toActivate;
    private Array<Entity> toDeactivate;
    private Array<Entity> toRemove;

    private ComponentBag[] components;
    private Array<EntitySystem> systems;

    private ListenerRegistry listenerRegistry;

    private Map<Class<? extends Component>, ComponentMapper> componentMappers;

    public Engine() {
        this.entities = new Array<>();
        this.toActivate = new Array<>();
        this.toDeactivate = new Array<>();
        this.toRemove = new Array<>();

        this.systems = new Array<>();
        this.listenerRegistry = new ListenerRegistry();
        this.componentMappers = new HashMap<>();
    }

    public void registerComponentClass(Class<? extends Component> clazz) {
        ComponentType.registerComponentType(clazz);
    }

    public void registerSystem(EntitySystem system) {
        this.systems.add(system);
    }

    public void registerListener(Listener listener) {
        listenerRegistry.registerListener(listener);
    }

    public void ready() {
        this.components = new ComponentBag[ComponentType.getRegisteredComponentTypeCount()];
        for (int i = 0; i < components.length; i++) {
            components[i] = new ComponentBag();
        }
    }

    public Array<Entity> getEntities() {
        return entities;
    }

    public int getEntityCount() {
        return entities.size;
    }

    public void update(float dt) {
        activatePending();
        deactivatePending();
        removePending();

        for (EntitySystem system : systems) {
            system.preProcessing();
            system.processEntities(dt);
            system.postProcessing();
        }
    }

    public Entity createEntity() {
        Entity entity = new Entity(this);
        entities.add(entity);
        return entity;
    }

    protected void removeEntity(Entity entity) {
        entity.deactivate();
        toRemove.add(entity);
    }

    private void removePending() {
        if (toRemove.isEmpty()) {
            return;
        }

        for (Entity entity : toRemove) {
            removeAllEntityComponents(entity.getId());
            entities.removeValue(entity, true);
        }

        toRemove.clear();
    }

    protected void activateEntity(Entity entity) {
        toActivate.add(entity);
    }

    private void activatePending() {
        if (toActivate.isEmpty()) {
            return;
        }

        for (Entity entity : toActivate) {
            for (EntitySystem system : systems) {
                if (system.interestedIn(entity)) {
                    system.entities.add(entity);
                }
            }
        }
        toActivate.clear();
    }

    protected void deactivateEntity(Entity entity) {
        toDeactivate.add(entity);
    }

    private void deactivatePending() {
        if (toDeactivate.isEmpty()) {
            return;
        }

        for (Entity entity : toDeactivate) {
            for (EntitySystem system : systems) {
                system.entities.removeValue(entity, true);
            }
        }

        toDeactivate.clear();
    }

    public void callEvent(Event event) {
        listenerRegistry.callEvent(event);
    }

    private void removeAllEntityComponents(int entityId) {
        for (int i = 0; i < components.length; i++) {
            components[i].insert(entityId, null);
        }
    }

    protected void addEntityComponent(Entity entity, Component component) {
        ComponentType type = ComponentType.getComponentType(component.getClass());
        components[type.getId()].insert(entity.getId(), component);
        entity.addComponentType(type);
    }


    protected void removeEntityComponent(Entity entity, Component component) {
        ComponentType type = ComponentType.getComponentType(component.getClass());
        components[type.getId()].remove(entity.getId());
        entity.removeComponentType(type);
    }

    protected <T extends Component> T getEntityComponent(Entity entity, Class<T> clazz) {
        ComponentType type = ComponentType.getComponentType(clazz);
        return clazz.cast(components[type.getId()].get(entity.getId()));
    }

    protected Component getEntityComponent(Entity entity, ComponentType type) {
        return components[type.getId()].get(entity.getId());
    }

    @SuppressWarnings("unchecked")
    public <T extends Component> ComponentMapper<T> getComponentMapper(Class<T> typeClass) {
        ComponentMapper<T> mapper = componentMappers.get(typeClass);
        if (mapper == null) {
            mapper = new ComponentMapper<>(this, typeClass);
            componentMappers.put(typeClass, mapper);
        }
        return mapper;
    }

}
